# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/02_ids_properties.ipynb.

# %% auto 0
__all__ = ['get_candidate_data', 'get_candidates', 'calc_candidate_duration', 'get_data_at_times', 'calc_rotation_angle',
           'calc_events_rotation_angle', 'calc_normal_direction', 'calc_events_normal_direction',
           'calc_events_vec_change', 'IDsPdPipeline', 'process_events']

# %% ../../notebooks/02_ids_properties.ipynb 1
#| code-summary: "Import all the packages needed for the project"
import polars as pl
import xarray as xr
from fastcore.all import patch

try:
    import modin.pandas as pd
    import modin.pandas as mpd
    from modin.config import ProgressBar
    ProgressBar.enable()
except ImportError:
    import pandas as pd
import pandas
    
import numpy as np

from datetime import timedelta

from loguru import logger

import pdpipe as pdp
from pdpipe.util import out_of_place_col_insert

# %% ../../notebooks/02_ids_properties.ipynb 2
def get_candidate_data(
    candidate: dict, data: xr.DataArray, neighbor: int = 0
) -> xr.DataArray:
    duration = candidate["tstop"] - candidate["tstart"]
    offset = neighbor * duration
    temp_tstart = candidate["tstart"] - offset
    temp_tstop = candidate["tstop"] + offset

    return data.sel(time=slice(temp_tstart, temp_tstop))


def get_candidates(candidates: pd.DataFrame, candidate_type=None, num: int = 4):
    if candidate_type is not None:
        _candidates = candidates[candidates["type"] == candidate_type]
    else:
        _candidates = candidates

    # Sample a specific number of candidates if num is provided and it's less than the total number
    if num < len(_candidates):
        logger.info(
            f"Sampling {num} {candidate_type} candidates out of {len(_candidates)}"
        )
        return _candidates.sample(num)
    else:
        return _candidates

# %% ../../notebooks/02_ids_properties.ipynb 4
from ..propeties.duration import calc_duration

# %% ../../notebooks/02_ids_properties.ipynb 5
def calc_candidate_duration(candidate, data, **kwargs):
    try:
        candidate_data = get_candidate_data(candidate, data)
        result = calc_duration(candidate_data, **kwargs)
        return pandas.Series(result)
    except Exception as e:
        logger.debug(
            f"Error for candidate {candidate} at {candidate['time']}: {str(e)}"
        )
        raise e

# %% ../../notebooks/02_ids_properties.ipynb 7
from ..propeties.mva import calc_candidate_mva_features

# %% ../../notebooks/02_ids_properties.ipynb 9
def get_data_at_times(data: xr.DataArray, times) -> np.ndarray:
    """
    Select data at specified times.
    """
    # Use xarray's selection capability if data supports it
    return data.sel(time=times, method="nearest").to_numpy()

# %% ../../notebooks/02_ids_properties.ipynb 10
def calc_rotation_angle(v1, v2):
    """
    Computes the rotation angle between two vectors.
    
    Parameters:
    - v1: The first vector(s).
    - v2: The second vector(s).
    """
    
    if v1.shape != v2.shape:
        raise ValueError("Vectors must have the same shape.")
    
    # Normalize the vectors
    v1_u = v1 / np.linalg.norm(v1, axis=-1, keepdims=True)
    v2_u = v2 / np.linalg.norm(v2, axis=-1, keepdims=True)
    
    # Calculate the cosine of the angle for each time step
    cosine_angle = np.sum(v1_u * v2_u, axis=-1)
    
    # Clip the values to handle potential floating point errors
    cosine_angle = np.clip(cosine_angle, -1, 1)
    
    angle = np.arccos(cosine_angle)
    
    # Convert the angles from radians to degrees
    return np.degrees(angle)


# %% ../../notebooks/02_ids_properties.ipynb 11
def calc_events_rotation_angle(events, data: xr.DataArray):
    """
    Computes the rotation angle(s) at two different time steps.
    """
    tstart = events['t.d_start'].to_numpy()
    tstop = events['t.d_end'].to_numpy()

    vecs_before = get_data_at_times(data, tstart)
    vecs_after = get_data_at_times(data, tstop)

    rotation_angles = calc_rotation_angle(vecs_before, vecs_after)
    return rotation_angles

# %% ../../notebooks/02_ids_properties.ipynb 13
def calc_normal_direction(v1, v2, normalize=True) -> np.ndarray:
    """
    Computes the normal direction of two vectors.

    Parameters
    ----------
    v1 : array_like 
        The first vector(s).
    v2 : array_like 
        The second vector(s).
    """
    c = np.cross(v1, v2)
    return c / np.linalg.norm(c, axis=-1, keepdims=True)


# %% ../../notebooks/02_ids_properties.ipynb 14
def calc_events_normal_direction(events, data: xr.DataArray):
    """
    Computes the normal directions(s) at two different time steps.
    """
    tstart = events['t.d_start'].to_numpy()
    tstop = events['t.d_end'].to_numpy()

    vecs_before = get_data_at_times(data, tstart)
    vecs_after = get_data_at_times(data, tstop)

    normal_directions = calc_normal_direction(vecs_before, vecs_after)
    # need to convert to list first, as only 1D array is supported
    return normal_directions.tolist()


# %% ../../notebooks/02_ids_properties.ipynb 15
def calc_events_vec_change(events, data: xr.DataArray):
    """
    Utils function to calculate features related to the change of the magnetic field
    """
    tstart = events['t.d_start'].to_numpy()
    tstop = events['t.d_end'].to_numpy()
    
    vecs_before = get_data_at_times(data, tstart)
    vecs_after = get_data_at_times(data, tstop)
    return (vecs_after - vecs_before).tolist()
    

# %% ../../notebooks/02_ids_properties.ipynb 18
@patch
def _transform(self: pdp.ApplyToRows, X, verbose):
    new_cols = X.apply(self._func, axis=1)
    if isinstance(new_cols, (pd.Series, pandas.Series)):
        loc = len(X.columns)
        if self._follow_column:
            loc = X.columns.get_loc(self._follow_column) + 1
        return out_of_place_col_insert(
            X=X, series=new_cols, loc=loc, column_name=self._colname
        )
    if isinstance(new_cols, (mpd.DataFrame, pandas.DataFrame)):
        sorted_cols = sorted(list(new_cols.columns))
        new_cols = new_cols[sorted_cols]
        if self._follow_column:
            inter_X = X
            loc = X.columns.get_loc(self._follow_column) + 1
            for colname in new_cols.columns:
                inter_X = out_of_place_col_insert(
                    X=inter_X,
                    series=new_cols[colname],
                    loc=loc,
                    column_name=colname,
                )
                loc += 1
            return inter_X
        assign_map = {
            colname: new_cols[colname] for colname in new_cols.columns
        }
        return X.assign(**assign_map)
    raise TypeError(  # pragma: no cover
        "Unexpected type generated by applying a function to a DataFrame."
        " Only Series and DataFrame are allowed."
    )

# %% ../../notebooks/02_ids_properties.ipynb 20
class IDsPdPipeline:
    @staticmethod
    def calc_duration(data: xr.DataArray, **kwargs):
        return pdp.ApplyToRows(
            lambda df: calc_candidate_duration(df, data, **kwargs),
            func_desc="calculating pre-duration parameters",
        )

    @staticmethod
    def calc_mva_features(data, **kwargs):
        return pdp.ApplyToRows(
            lambda df: calc_candidate_mva_features(df, data, **kwargs),
            func_desc="calculating MVA features",
        )

    @staticmethod
    def calc_vec_change(data, **kwargs):
        return pdp.ColByFrameFunc(
            "dB",
            lambda df: calc_events_vec_change(df, data, **kwargs),
            func_desc="calculating compound change",
        )

    @staticmethod
    def calc_rotation_angle(data, **kwargs):
        return pdp.ColByFrameFunc(
            "rotation_angle",
            lambda df: calc_events_rotation_angle(df, data, **kwargs),
            func_desc="calculating rotation angle",
        )

    @staticmethod
    def calc_normal_direction(data, name="normal_direction", **kwargs):
        return pdp.ColByFrameFunc(
            name,
            lambda df: calc_events_normal_direction(df, data, **kwargs),
            func_desc="calculating normal direction",
        )

# %% ../../notebooks/02_ids_properties.ipynb 21
from beforerr.polars import convert_to_pd_dataframe, decompose_vector

# %% ../../notebooks/02_ids_properties.ipynb 22
from typing import Literal


def process_events(
    candidates_pl: pl.DataFrame,  # potential candidates DataFrame
    sat_fgm: xr.DataArray,  # satellite FGM data
    data_resolution: timedelta,  # time resolution of the data
    modin=True,
    method: Literal["fit", "derivative"]= "fit",
    **kwargs,
) -> pl.DataFrame:
    "Process candidates DataFrame"

    candidates = pd.DataFrame(convert_to_pd_dataframe(candidates_pl, modin=modin))
    
    if method == "fit":
        duration_method = "distance"  
    else:
        duration_method = "derivative"

    candidates = (
        IDsPdPipeline.calc_duration(sat_fgm, method=duration_method, **kwargs)
        .apply(candidates)
        .dropna()
    )  # Remove candidates with NaN values)

    ids = (
        IDsPdPipeline.calc_mva_features(sat_fgm, method=method, **kwargs)
        + IDsPdPipeline.calc_vec_change(sat_fgm)
        + IDsPdPipeline.calc_rotation_angle(sat_fgm)
        + IDsPdPipeline.calc_normal_direction(sat_fgm, name="k")
    ).apply(candidates)

    if isinstance(ids, mpd.DataFrame):
        ids = ids._to_pandas()

    vectors2decompose = ["dB", "dB_lmn", "k", "Vl", "Vn"]

    df = pl.DataFrame(
        ids.dropna(), schema_overrides={vec: pl.List for vec in vectors2decompose}
    )  # ArrowInvalid: Could not convert [0.9799027968348948, -0.17761542644940076, -0.07309766783111293] with type list: tried to convert to double

    if method == "fit":
        duration_expr = pl.col("fit.vars.sigma") * 2
    else:
        duration_expr = (
            pl.col("t.d_end") - pl.col("t.d_start")
        ).dt.total_nanoseconds() / 1e9  # convert to seconds

    for vec in vectors2decompose:
        df = decompose_vector(df, vec)

    return df.with_columns(duration=duration_expr).drop(vectors2decompose)
    # ValueError: Data type fixed_size_list[pyarrow] not supported by interchange protocol



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/utils/01_plotting.ipynb.

# %% auto 0
__all__ = ['time_stamp', 'setup_mva_plot', 'plot_candidate']

# %% ../../notebooks/utils/01_plotting.ipynb 0
import matplotlib.pyplot as plt

import xarray as xr
import pandas as pd
from datetime import datetime

from pyspedas.cotrans.minvar_matrix_make import minvar_matrix_make
from pyspedas import tvector_rotate
from pyspedas.analysis.tvectot import tvectot

from pytplot import tplot
from pytplot import store_data
from pytplot import timebar, degap, options

from matplotlib.pyplot import Axes

# %% ../../notebooks/utils/01_plotting.ipynb 2
try:
    import scienceplots
    plt.style.use(['science', 'nature', 'notebook'])
except ImportError:
    pass

# %% ../../notebooks/utils/01_plotting.ipynb 3
def time_stamp(ts):
    "Return POSIX timestamp as float."
    return pd.Timestamp(ts, tz="UTC").timestamp()


def setup_mva_plot(
    data: xr.DataArray,
    tstart: datetime,
    tstop: datetime,
    mva_tstart: datetime = None,
    mva_tstop: datetime = None,
):
    if mva_tstart is None:
        mva_tstart = tstart
    if mva_tstop is None:
        mva_tstop = tstop

    mva_b = data.sel(time=slice(mva_tstart, mva_tstop))
    store_data("fgm", data={"x": mva_b.time, "y": mva_b})
    minvar_matrix_make("fgm")  # get the MVA matrix

    temp_b = data.sel(time=slice(tstart, tstop))
    store_data("fgm", data={"x": temp_b.time, "y": temp_b})
    tvar = tvector_rotate("fgm_mva_mat", "fgm")[0]
    ysubtitle = "[nT LMN]"
    legend_names = [r"$B_l$", r"$B_m$", r"$B_n$"]
    
    tvar2plot = tvectot(tvar, join_component=True)
    legend_names = legend_names + [r"$B_{total}$"]

    options(tvar2plot, "ytitle", "$B$")
    options(tvar2plot, "ysubtitle", ysubtitle)
    options(tvar2plot, "legend_names", legend_names)

    degap(tvar2plot)
    return tvar2plot

# %% ../../notebooks/utils/01_plotting.ipynb 5
from pyspedas.analysis.deriv_data import deriv_data
from ..integration import J_FACTOR
import pytplot
from pytplot import split_vec

def plot_candidate(
    event: dict,
    data: xr.DataArray,
    add_ids_properties=True,
    plot_current_density=False,
    plot_fit_data=False,
    add_timebars=True,
    add_plasma_params=False,
    **kwargs,
):
    if pd.notnull(event.get("t.d_start")) and pd.notnull(event.get("t.d_end")):
        tvar = setup_mva_plot(
            data,
            event["tstart"],
            event["tstop"],
            event["t.d_start"],
            event["t.d_end"],
        )
    else:
        tvar = setup_mva_plot(data, event["tstart"], event["tstop"])
    
    tvars2plot = [tvar]

    if plot_current_density:
        Bl = split_vec(tvar)[0]
        dBldt = deriv_data(Bl)[0]
        v_k = event.get("v_k")
        pytplot.data_quants[dBldt] = pytplot.data_quants[dBldt] / v_k * J_FACTOR.value
        tvars2plot.append(dBldt)
        
        options(dBldt, "ytitle", "$J$")
        options(dBldt, "ysubtitle", "[nA/m$^2$]")
        options(dBldt, "legend_names", "$J_m$")

    if add_timebars:
        d_time = event.get("t.d_time")
        d_start = event.get("t.d_start")
        d_stop = event.get("t.d_end")

        if d_time:
            timebar(time_stamp(d_time), color="red")
        if d_start and pd.notnull(d_start):
            timebar(time_stamp(d_start))
        if d_stop and pd.notnull(d_stop):
            timebar(time_stamp(d_stop))

    title = ""
    
    if add_ids_properties:
        thickness = event.get("L_k")
        current_density = event.get("j0_k")
        title += "#Discontinuity properties# "
        if thickness:
            title += rf"$L: {thickness:.2f} \mathrm{{km}}$"
        if current_density:
            title += rf", $j: {current_density:.2f} \mathrm{{nA/m}}^2$"
    
    if add_plasma_params:
        plasma_speed = event.get("plasma_speed")
        plasma_density = event.get("plasma_density")
        plasma_temperature = event.get("plasma_temperature")

        title += "\n#Plasma parameters# "
        if plasma_speed:
            title += rf"$V_i: {plasma_speed:.2f} \mathrm{{km/s}}$"
        if plasma_density:
            title += rf", $n_i: {plasma_density:.2f} \mathrm{{cm}}^{{-3}}$"
        if plasma_temperature:
            title += rf", $T_i: {plasma_temperature:.2f} \mathrm{{eV}}$"

        # options(tvar, "title", title)
        
    for tvar2plot in tvars2plot:
        options(tvar2plot, "thick", 2)
        options(tvar2plot, "char_size", 16)

    fig, axes = tplot(tvars2plot, return_plot_objects=True)
    if isinstance(axes, Axes):
        axes = [axes]

    base_axis = axes[0]

    if plot_fit_data:
        fit_data = event.get("fit.best_fit")
        fit_time = event.get("fit.time")

        c = event.get("fit.vars.c")
        amp = event.get("fit.vars.amplitude")
        sigma = event.get('fit.vars.sigma')
        
        d_star = event.get("d_star")
        rsquared = event.get("fit.stat.rsquared")
        chisqr = event.get("fit.stat.chisqr")
        
        base_axis.plot(d_time, c + amp / 2, marker="o", markersize=10, color="red")
        if fit_time is not None and fit_data is not None:
            base_axis.plot(fit_time, fit_data, label="Fit", color="black", linestyle="--")
        
        title += f"\n#Fit# $\max dB/dt$: {d_star:.2f}, $R^2$: {rsquared:.2f}, $\chi^2$: {chisqr:.2f}"
        title += f"\n#Fit# $c$: {c:.2f}, $Amp$: {amp:.2f}, $\Sigma$: {sigma:.2f}"

    # add title to the first plot
    base_axis.set_title(title)

    return fig, axes



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/properties/00_mva.ipynb.

# %% auto 0
__all__ = ['minvar', 'calc_mva_features', 'calc_maxiumum_variance_direction', 'fit_maxiumum_variance_direction',
           'calc_candidate_mva_features']

# %% ../../notebooks/properties/00_mva.ipynb 1
import xarray as xr
import numpy as np
import pandas as pd

from lmfit.models import StepModel, ConstantModel
from lmfit import Parameters

from typing import Literal

# %% ../../notebooks/properties/00_mva.ipynb 3
def minvar(data: np.ndarray):
    """
    see `pyspedas.cotrans.minvar`

    This program computes the principal variance directions and variances of a
    vector quantity as well as the associated eigenvalues.

    Parameters
    -----------
    data:
        Vxyz, an (npoints, ndim) array of data(ie Nx3)

    Returns
    -------
    vrot:
        an array of (npoints, ndim) containing the rotated data in the new coordinate system, ijk.
        Vi(maximum direction)=vrot[0,:]
        Vj(intermediate direction)=vrot[1,:]
        Vk(minimum variance direction)=Vrot[2,:]
    v:
        an (ndim,ndim) array containing the principal axes vectors
        Maximum variance direction eigenvector, Vi=v[*,0]
        Intermediate variance direction, Vj=v[*,1] (descending order)
    w:
        the eigenvalues of the computation
    """

    #  Min var starts here
    # data must be Nx3
    vecavg = np.nanmean(np.nan_to_num(data, nan=0.0), axis=0)

    mvamat = np.zeros((3, 3))
    for i in range(3):
        for j in range(3):
            mvamat[i, j] = (
                np.nanmean(np.nan_to_num(data[:, i] * data[:, j], nan=0.0))
                - vecavg[i] * vecavg[j]
            )

    # Calculate eigenvalues and eigenvectors
    w, v = np.linalg.eigh(mvamat, UPLO="U")

    # Sorting to ensure descending order
    w = np.abs(w)
    idx = np.flip(np.argsort(w))

    # IDL compatability
    if True:
        if np.sum(w) == 0.0:
            idx = [0, 2, 1]

    w = w[idx]
    v = v[:, idx]

    # Rotate intermediate var direction if system is not Right Handed
    YcrossZdotX = v[0, 0] * (v[1, 1] * v[2, 2] - v[2, 1] * v[1, 2])
    if YcrossZdotX < 0:
        v[:, 1] = -v[:, 1]
        # v[:, 2] = -v[:, 2] # Should not it is being flipped at Z-axis?

    # Ensure minvar direction is along +Z (for FAC system)
    if v[2, 2] < 0:
        v[:, 2] = -v[:, 2]
        v[:, 1] = -v[:, 1]

    vrot = np.array([np.dot(row, v) for row in data])

    return vrot, v, w

# %% ../../notebooks/properties/00_mva.ipynb 5
def calc_mva_features(data: np.ndarray):
    """
    Compute MVA features based on the given data array.

    Parameters:
    - data (np.ndarray): Input data

    Returns:
    - List: Computed features
    """

    # Compute variance properties
    vrot, v, eigs = minvar(data)

    # Maximum variance direction eigenvector
    Vl = v[:, 0]
    Vn = v[:, 2]

    vec_mag = np.linalg.norm(vrot, axis=1)

    # Compute changes in each component of B_rot
    dvec = vrot[0] - vrot[-1]

    # Compute mean values
    vec_mag_mean = np.mean(vec_mag)
    vec_n_mean = np.mean(vrot[:, 2])
    VnOverVmag = vec_n_mean / vec_mag_mean

    # Compute relative changes in magnitude
    dvec_mag = vec_mag[-1] - vec_mag[0]
    dBOverB = np.abs(dvec_mag / vec_mag_mean)
    dBOverB_max = (np.max(vec_mag) - np.min(vec_mag)) / vec_mag_mean
    
    result = {
        "Vl": Vl,
        "Vn": Vn,
        "b_mag": vec_mag_mean,
        "b_n": vec_n_mean,
        'B.vec.before': vrot[0],
        'B.vec.after': vrot[-1],
        'B.before': vec_mag[0],
        'B.after': vec_mag[-1],
        "db_mag": dvec_mag,
        "bn_over_b": VnOverVmag,
        "db_over_b": dBOverB,
        "db_over_b_max": dBOverB_max,
        "dB_lmn": dvec,
    }
    
    result = pd.Series(result)
    return result, vrot

# %% ../../notebooks/properties/00_mva.ipynb 6
def calc_maxiumum_variance_direction(data: xr.DataArray, datetime_unit="s", **kwargs):
    d_data = data.differentiate("time", datetime_unit=datetime_unit)
    return pd.Series({"d_star": abs(d_data).max(dim="time").item()})

# %% ../../notebooks/properties/00_mva.ipynb 9
def fit_maxiumum_variance_direction(
    data: xr.DataArray, datetime_unit="s", return_best_fit: bool = True, **kwargs
):
    """
    Fit maximum variance direction data by model

    Note:
        - see `datetime_to_numeric` in `xarray.core.duck_array_ops` for more details about converting datetime to numeric
        - Xarray uses the numpy dtypes datetime64[ns] and timedelta64[ns] to represent datetime data.
    """
    time = data["time"].values
    x = (time - min(time)) / np.timedelta64(1, datetime_unit)
    y = data.values

    x_min, x_max = min(x), max(x)
    x_width = x_max - x_min

    # Create a model
    step_mod = StepModel(form="logistic")
    const_mod = ConstantModel()
    mod = step_mod + const_mod

    # Create parameters
    params = Parameters()

    init_amplitude = y[-1] - y[0]

    params.add(
        "center",
        value=(x_max + x_min) / 2.0,
        # min=x_min + x_width / 7.0,
        # max=x_max - x_width / 7.0,
    )
    params.add(
        "amplitude",
        value=init_amplitude,
        max=abs(init_amplitude) * 2.0,
        min=-abs(init_amplitude) * 2.0,
    )
    params.add("sigma", value=x_width / 7.0, min=0)

    if True:
        int_center_y = (y[-1] + y[0]) / 2.0
        c1 = 1 / 8
        params.add(
            "center_y",
            min=int_center_y - c1 * np.abs(init_amplitude),
            max=int_center_y + c1 * np.abs(init_amplitude),
        )
        params.add("c", expr="center_y - amplitude / 2.0")

    # Ensure there are enough data points to fit the model
    if len(y) < 4:
        # Not enough data points, return None or an empty result instead of raising an error
        amplitude = np.nan
        sigma = np.nan
        center = np.nan
        rsquared = np.nan
        chisqr = np.nan
        c = np.nan
        best_fit = np.nan
    else:
        out = mod.fit(y, params, x=x)
        amplitude = out.params["amplitude"].value
        sigma = out.params["sigma"].value
        center = out.params["center"].value
        c = out.params["c"].value
        rsquared = out.rsquared
        chisqr = out.chisqr
        best_fit = out.best_fit

    max_df = amplitude / (4 * sigma)

    d_time = min(time) + center * np.timedelta64(1, datetime_unit).astype(
        "timedelta64[ns]"
    )

    result = pd.Series(
        {
            "fit.vars.amplitude": amplitude,
            "fit.vars.sigma": sigma,
            "t.d_time": d_time,
            "d_star": max_df,
            "fit.vars.c": c,
            "fit.stat.rsquared": rsquared,
            "fit.stat.chisqr": chisqr,
        }
    )
    if return_best_fit:
        result["fit.best_fit"] = best_fit
        result["fit.time"] = time
    return result

# %% ../../notebooks/properties/00_mva.ipynb 10
def calc_candidate_mva_features(
    event, data: xr.DataArray, method=Literal["fit", "derivative"], **kwargs
):
    event_data = data.sel(time=slice(event["t.d_start"], event["t.d_end"]))

    mva_features, vrot = calc_mva_features(event_data.to_numpy())

    event_data_l = xr.DataArray(
        vrot[:, 0], dims=["time"], coords={"time": event_data.time}
    )

    if method == "fit":
        result = fit_maxiumum_variance_direction(event_data_l, **kwargs)
    elif method == "derivative":
        result = calc_maxiumum_variance_direction(event_data_l, **kwargs)
    return pd.concat([mva_features, result])



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/ids_finder',
                'doc_host': 'https://Beforerr.github.io',
                'git_url': 'https://github.com/Beforerr/ids_finder',
                'lib_path': 'discontinuitypy'},
  'syms': { 'discontinuitypy.config': { 'discontinuitypy.config.IDsConfig': ('ids_config.html#idsconfig', 'discontinuitypy/config.py'),
                                        'discontinuitypy.config.IDsConfig.export': ( 'ids_config.html#idsconfig.export',
                                                                                     'discontinuitypy/config.py'),
                                        'discontinuitypy.config.IDsConfig.fname': ( 'ids_config.html#idsconfig.fname',
                                                                                    'discontinuitypy/config.py'),
                                        'discontinuitypy.config.IDsConfig.load': ( 'ids_config.html#idsconfig.load',
                                                                                   'discontinuitypy/config.py'),
                                        'discontinuitypy.config.IDsConfig.path': ( 'ids_config.html#idsconfig.path',
                                                                                   'discontinuitypy/config.py'),
                                        'discontinuitypy.config.IDsConfig.timeranges': ( 'ids_config.html#idsconfig.timeranges',
                                                                                         'discontinuitypy/config.py'),
                                        'discontinuitypy.config.SpeasyIDsConfig': ( 'ids_config.html#speasyidsconfig',
                                                                                    'discontinuitypy/config.py'),
                                        'discontinuitypy.config.SpeasyIDsConfig._get_and_process_data': ( 'ids_config.html#speasyidsconfig._get_and_process_data',
                                                                                                          'discontinuitypy/config.py'),
                                        'discontinuitypy.config.SpeasyIDsConfig.e_temp_var': ( 'ids_config.html#speasyidsconfig.e_temp_var',
                                                                                               'discontinuitypy/config.py'),
                                        'discontinuitypy.config.SpeasyIDsConfig.get_and_process_data': ( 'ids_config.html#speasyidsconfig.get_and_process_data',
                                                                                                         'discontinuitypy/config.py'),
                                        'discontinuitypy.config.SpeasyIDsConfig.get_vars': ( 'ids_config.html#speasyidsconfig.get_vars',
                                                                                             'discontinuitypy/config.py'),
                                        'discontinuitypy.config.SpeasyIDsConfig.get_vars_df': ( 'ids_config.html#speasyidsconfig.get_vars_df',
                                                                                                'discontinuitypy/config.py'),
                                        'discontinuitypy.config.SpeasyIDsConfig.ion_temp_var': ( 'ids_config.html#speasyidsconfig.ion_temp_var',
                                                                                                 'discontinuitypy/config.py'),
                                        'discontinuitypy.config.SpeasyIDsConfig.mag_vars': ( 'ids_config.html#speasyidsconfig.mag_vars',
                                                                                             'discontinuitypy/config.py'),
                                        'discontinuitypy.config.SpeasyIDsConfig.model_post_init': ( 'ids_config.html#speasyidsconfig.model_post_init',
                                                                                                    'discontinuitypy/config.py'),
                                        'discontinuitypy.config.SpeasyIDsConfig.plasma_vars': ( 'ids_config.html#speasyidsconfig.plasma_vars',
                                                                                                'discontinuitypy/config.py'),
                                        'discontinuitypy.config.SpeasyIDsConfig.set_data_from_vars': ( 'ids_config.html#speasyidsconfig.set_data_from_vars',
                                                                                                       'discontinuitypy/config.py'),
                                        'discontinuitypy.config.standardize_plasma_data': ( 'ids_config.html#standardize_plasma_data',
                                                                                            'discontinuitypy/config.py')},
            'discontinuitypy.core.detection': { 'discontinuitypy.core.detection.detect_events': ( 'ids_detection.html#detect_events',
                                                                                                  'discontinuitypy/core/detection.py'),
                                                'discontinuitypy.core.detection.pl_format_time': ( 'ids_detection.html#pl_format_time',
                                                                                                   'discontinuitypy/core/detection.py')},
            'discontinuitypy.core.event': { 'discontinuitypy.core.event.get_event_data': ( 'event.html#get_event_data',
                                                                                           'discontinuitypy/core/event.py')},
            'discontinuitypy.core.pipeline': { 'discontinuitypy.core.pipeline.compress_data_by_events': ( 'ids_finder.html#compress_data_by_events',
                                                                                                          'discontinuitypy/core/pipeline.py'),
                                               'discontinuitypy.core.pipeline.compress_data_by_intervals': ( 'ids_finder.html#compress_data_by_intervals',
                                                                                                             'discontinuitypy/core/pipeline.py'),
                                               'discontinuitypy.core.pipeline.compress_dfs_by_intervals': ( 'ids_finder.html#compress_dfs_by_intervals',
                                                                                                            'discontinuitypy/core/pipeline.py'),
                                               'discontinuitypy.core.pipeline.extract_features': ( 'ids_finder.html#extract_features',
                                                                                                   'discontinuitypy/core/pipeline.py'),
                                               'discontinuitypy.core.pipeline.ids_finder': ( 'ids_finder.html#ids_finder',
                                                                                             'discontinuitypy/core/pipeline.py')},
            'discontinuitypy.core.propeties': { 'discontinuitypy.core.propeties.IDsPdPipeline': ( 'ids_properties.html#idspdpipeline',
                                                                                                  'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.IDsPdPipeline.calc_duration': ( 'ids_properties.html#idspdpipeline.calc_duration',
                                                                                                                'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.IDsPdPipeline.calc_mva_features': ( 'ids_properties.html#idspdpipeline.calc_mva_features',
                                                                                                                    'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.IDsPdPipeline.calc_normal_direction': ( 'ids_properties.html#idspdpipeline.calc_normal_direction',
                                                                                                                        'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.IDsPdPipeline.calc_rotation_angle': ( 'ids_properties.html#idspdpipeline.calc_rotation_angle',
                                                                                                                      'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.IDsPdPipeline.calc_vec_change': ( 'ids_properties.html#idspdpipeline.calc_vec_change',
                                                                                                                  'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.calc_candidate_duration': ( 'ids_properties.html#calc_candidate_duration',
                                                                                                            'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.calc_events_normal_direction': ( 'ids_properties.html#calc_events_normal_direction',
                                                                                                                 'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.calc_events_rotation_angle': ( 'ids_properties.html#calc_events_rotation_angle',
                                                                                                               'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.calc_events_vec_change': ( 'ids_properties.html#calc_events_vec_change',
                                                                                                           'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.calc_normal_direction': ( 'ids_properties.html#calc_normal_direction',
                                                                                                          'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.calc_rotation_angle': ( 'ids_properties.html#calc_rotation_angle',
                                                                                                        'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.get_candidate_data': ( 'ids_properties.html#get_candidate_data',
                                                                                                       'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.get_candidates': ( 'ids_properties.html#get_candidates',
                                                                                                   'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.get_data_at_times': ( 'ids_properties.html#get_data_at_times',
                                                                                                      'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.pdp.ApplyToRows._transform': ( 'ids_properties.html#pdp.applytorows._transform',
                                                                                                               'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.process_events': ( 'ids_properties.html#process_events',
                                                                                                   'discontinuitypy/core/propeties.py')},
            'discontinuitypy.datasets': { 'discontinuitypy.datasets.IDsDataset': ( 'datasets.html#idsdataset',
                                                                                   'discontinuitypy/datasets.py'),
                                          'discontinuitypy.datasets.IDsDataset.plot': ( 'datasets.html#idsdataset.plot',
                                                                                        'discontinuitypy/datasets.py'),
                                          'discontinuitypy.datasets.IDsDataset.plot_candidate': ( 'datasets.html#idsdataset.plot_candidate',
                                                                                                  'discontinuitypy/datasets.py'),
                                          'discontinuitypy.datasets.IDsDataset.plot_candidates': ( 'datasets.html#idsdataset.plot_candidates',
                                                                                                   'discontinuitypy/datasets.py'),
                                          'discontinuitypy.datasets.IDsDataset.update_events': ( 'datasets.html#idsdataset.update_events',
                                                                                                 'discontinuitypy/datasets.py'),
                                          'discontinuitypy.datasets.IDsDataset.update_events_with_plasma_data': ( 'datasets.html#idsdataset.update_events_with_plasma_data',
                                                                                                                  'discontinuitypy/datasets.py'),
                                          'discontinuitypy.datasets.IDsDataset.update_events_with_temp_data': ( 'datasets.html#idsdataset.update_events_with_temp_data',
                                                                                                                'discontinuitypy/datasets.py'),
                                          'discontinuitypy.datasets.IdsEvents': ('datasets.html#idsevents', 'discontinuitypy/datasets.py'),
                                          'discontinuitypy.datasets.IdsEvents.Config': ( 'datasets.html#idsevents.config',
                                                                                         'discontinuitypy/datasets.py'),
                                          'discontinuitypy.datasets.IdsEvents.export': ( 'datasets.html#idsevents.export',
                                                                                         'discontinuitypy/datasets.py'),
                                          'discontinuitypy.datasets.IdsEvents.find_events': ( 'datasets.html#idsevents.find_events',
                                                                                              'discontinuitypy/datasets.py'),
                                          'discontinuitypy.datasets.IdsEvents.get_event': ( 'datasets.html#idsevents.get_event',
                                                                                            'discontinuitypy/datasets.py'),
                                          'discontinuitypy.datasets.IdsEvents.get_event_data': ( 'datasets.html#idsevents.get_event_data',
                                                                                                 'discontinuitypy/datasets.py'),
                                          'discontinuitypy.datasets.IdsEvents.get_events': ( 'datasets.html#idsevents.get_events',
                                                                                             'discontinuitypy/datasets.py'),
                                          'discontinuitypy.datasets.log_event_change': ( 'datasets.html#log_event_change',
                                                                                         'discontinuitypy/datasets.py'),
                                          'discontinuitypy.datasets.write': ('datasets.html#write', 'discontinuitypy/datasets.py')},
            'discontinuitypy.datasets_kedro': { 'discontinuitypy.datasets_kedro.E1IDsDataset': ( 'datasets_kedro.html#e1idsdataset',
                                                                                                 'discontinuitypy/datasets_kedro.py'),
                                                'discontinuitypy.datasets_kedro.cIDsDataset': ( 'datasets_kedro.html#cidsdataset',
                                                                                                'discontinuitypy/datasets_kedro.py'),
                                                'discontinuitypy.datasets_kedro.cIDsDataset.__init__': ( 'datasets_kedro.html#cidsdataset.__init__',
                                                                                                         'discontinuitypy/datasets_kedro.py'),
                                                'discontinuitypy.datasets_kedro.cIDsDataset.load_data': ( 'datasets_kedro.html#cidsdataset.load_data',
                                                                                                          'discontinuitypy/datasets_kedro.py'),
                                                'discontinuitypy.datasets_kedro.cIDsDataset.load_events': ( 'datasets_kedro.html#cidsdataset.load_events',
                                                                                                            'discontinuitypy/datasets_kedro.py')},
            'discontinuitypy.detection.variance': { 'discontinuitypy.detection.variance.add_neighbor_std': ( 'detection/variance.html#add_neighbor_std',
                                                                                                             'discontinuitypy/detection/variance.py'),
                                                    'discontinuitypy.detection.variance.compute_combinded_std': ( 'detection/variance.html#compute_combinded_std',
                                                                                                                  'discontinuitypy/detection/variance.py'),
                                                    'discontinuitypy.detection.variance.compute_index_diff': ( 'detection/variance.html#compute_index_diff',
                                                                                                               'discontinuitypy/detection/variance.py'),
                                                    'discontinuitypy.detection.variance.compute_index_fluctuation': ( 'detection/variance.html#compute_index_fluctuation',
                                                                                                                      'discontinuitypy/detection/variance.py'),
                                                    'discontinuitypy.detection.variance.compute_index_std': ( 'detection/variance.html#compute_index_std',
                                                                                                              'discontinuitypy/detection/variance.py'),
                                                    'discontinuitypy.detection.variance.compute_indices': ( 'detection/variance.html#compute_indices',
                                                                                                            'discontinuitypy/detection/variance.py'),
                                                    'discontinuitypy.detection.variance.compute_std': ( 'detection/variance.html#compute_std',
                                                                                                        'discontinuitypy/detection/variance.py'),
                                                    'discontinuitypy.detection.variance.filter_indices': ( 'detection/variance.html#filter_indices',
                                                                                                           'discontinuitypy/detection/variance.py'),
                                                    'discontinuitypy.detection.variance.pl_dvec': ( 'detection/variance.html#pl_dvec',
                                                                                                    'discontinuitypy/detection/variance.py')},
            'discontinuitypy.integration': { 'discontinuitypy.integration.calc_combined_features': ( 'mag_plasma.html#calc_combined_features',
                                                                                                     'discontinuitypy/integration.py'),
                                             'discontinuitypy.integration.calc_plasma_parameter_change': ( 'mag_plasma.html#calc_plasma_parameter_change',
                                                                                                           'discontinuitypy/integration.py'),
                                             'discontinuitypy.integration.calc_rotation_angle_pl': ( 'mag_plasma.html#calc_rotation_angle_pl',
                                                                                                     'discontinuitypy/integration.py'),
                                             'discontinuitypy.integration.combine_features': ( 'mag_plasma.html#combine_features',
                                                                                               'discontinuitypy/integration.py'),
                                             'discontinuitypy.integration.compute_Alfven_current': ( 'mag_plasma.html#compute_alfven_current',
                                                                                                     'discontinuitypy/integration.py'),
                                             'discontinuitypy.integration.compute_inertial_length': ( 'mag_plasma.html#compute_inertial_length',
                                                                                                      'discontinuitypy/integration.py'),
                                             'discontinuitypy.integration.format_time': ( 'mag_plasma.html#format_time',
                                                                                          'discontinuitypy/integration.py'),
                                             'discontinuitypy.integration.interpolate': ( 'mag_plasma.html#interpolate',
                                                                                          'discontinuitypy/integration.py'),
                                             'discontinuitypy.integration.interpolate2': ( 'mag_plasma.html#interpolate2',
                                                                                           'discontinuitypy/integration.py'),
                                             'discontinuitypy.integration.vector_project': ( 'mag_plasma.html#vector_project',
                                                                                             'discontinuitypy/integration.py'),
                                             'discontinuitypy.integration.vector_project_pl': ( 'mag_plasma.html#vector_project_pl',
                                                                                                'discontinuitypy/integration.py')},
            'discontinuitypy.missions': { 'discontinuitypy.missions.WindConfig': ( 'mission_config.html#windconfig',
                                                                                   'discontinuitypy/missions.py'),
                                          'discontinuitypy.missions.WindMeta': ( 'mission_config.html#windmeta',
                                                                                 'discontinuitypy/missions.py')},
            'discontinuitypy.propeties.duration': { 'discontinuitypy.propeties.duration.calc_duration': ( 'properties/duration.html#calc_duration',
                                                                                                          'discontinuitypy/propeties/duration.py'),
                                                    'discontinuitypy.propeties.duration.find_start_end_times': ( 'properties/duration.html#find_start_end_times',
                                                                                                                 'discontinuitypy/propeties/duration.py'),
                                                    'discontinuitypy.propeties.duration.get_time_from_condition': ( 'properties/duration.html#get_time_from_condition',
                                                                                                                    'discontinuitypy/propeties/duration.py'),
                                                    'discontinuitypy.propeties.duration.ts_max_derivative': ( 'properties/duration.html#ts_max_derivative',
                                                                                                              'discontinuitypy/propeties/duration.py'),
                                                    'discontinuitypy.propeties.duration.ts_max_distance': ( 'properties/duration.html#ts_max_distance',
                                                                                                            'discontinuitypy/propeties/duration.py')},
            'discontinuitypy.propeties.mva': { 'discontinuitypy.propeties.mva.calc_candidate_mva_features': ( 'properties/mva.html#calc_candidate_mva_features',
                                                                                                              'discontinuitypy/propeties/mva.py'),
                                               'discontinuitypy.propeties.mva.calc_maxiumum_variance_direction': ( 'properties/mva.html#calc_maxiumum_variance_direction',
                                                                                                                   'discontinuitypy/propeties/mva.py'),
                                               'discontinuitypy.propeties.mva.calc_mva_features': ( 'properties/mva.html#calc_mva_features',
                                                                                                    'discontinuitypy/propeties/mva.py'),
                                               'discontinuitypy.propeties.mva.fit_maxiumum_variance_direction': ( 'properties/mva.html#fit_maxiumum_variance_direction',
                                                                                                                  'discontinuitypy/propeties/mva.py'),
                                               'discontinuitypy.propeties.mva.minvar': ( 'properties/mva.html#minvar',
                                                                                         'discontinuitypy/propeties/mva.py')},
            'discontinuitypy.utils.analysis': { 'discontinuitypy.utils.analysis.filter_before_jupiter': ( 'utils/analysis_utils.html#filter_before_jupiter',
                                                                                                          'discontinuitypy/utils/analysis.py'),
                                                'discontinuitypy.utils.analysis.filter_tranges_ds': ( 'utils/analysis_utils.html#filter_tranges_ds',
                                                                                                      'discontinuitypy/utils/analysis.py'),
                                                'discontinuitypy.utils.analysis.link_coord2dim': ( 'utils/analysis_utils.html#link_coord2dim',
                                                                                                   'discontinuitypy/utils/analysis.py'),
                                                'discontinuitypy.utils.analysis.n2_normalize': ( 'utils/analysis_utils.html#n2_normalize',
                                                                                                 'discontinuitypy/utils/analysis.py')},
            'discontinuitypy.utils.basic': { 'discontinuitypy.utils.basic._expand_selectors': ( 'utils/basic.html#_expand_selectors',
                                                                                                'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.calc_vec_mag': ( 'utils/basic.html#calc_vec_mag',
                                                                                           'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.check_fgm': ( 'utils/basic.html#check_fgm',
                                                                                        'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.concat_df': ( 'utils/basic.html#concat_df',
                                                                                        'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.concat_partitions': ( 'utils/basic.html#concat_partitions',
                                                                                                'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.df2ts': ( 'utils/basic.html#df2ts',
                                                                                    'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.filter_tranges': ( 'utils/basic.html#filter_tranges',
                                                                                             'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.filter_tranges_df': ( 'utils/basic.html#filter_tranges_df',
                                                                                                'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.format_timedelta': ( 'utils/basic.html#format_timedelta',
                                                                                               'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.partition_data_by_time': ( 'utils/basic.html#partition_data_by_time',
                                                                                                     'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.partition_data_by_ts': ( 'utils/basic.html#partition_data_by_ts',
                                                                                                   'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.partition_data_by_year': ( 'utils/basic.html#partition_data_by_year',
                                                                                                     'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.partition_data_by_year_month': ( 'utils/basic.html#partition_data_by_year_month',
                                                                                                           'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.pl.DataFrame.plot': ( 'utils/basic.html#pl.dataframe.plot',
                                                                                                'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.resample': ( 'utils/basic.html#resample',
                                                                                       'discontinuitypy/utils/basic.py')},
            'discontinuitypy.utils.plot': { 'discontinuitypy.utils.plot.plot_candidate': ( 'utils/plotting.html#plot_candidate',
                                                                                           'discontinuitypy/utils/plot.py'),
                                            'discontinuitypy.utils.plot.setup_mva_plot': ( 'utils/plotting.html#setup_mva_plot',
                                                                                           'discontinuitypy/utils/plot.py'),
                                            'discontinuitypy.utils.plot.time_stamp': ( 'utils/plotting.html#time_stamp',
                                                                                       'discontinuitypy/utils/plot.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/detection/01_variance.ipynb.

# %% auto 0
__all__ = ['INDEX_STD_THRESHOLD', 'INDEX_FLUC_THRESHOLD', 'INDEX_DIFF_THRESHOLD', 'SPARSE_THRESHOLD', 'compute_std',
           'add_neighbor_std', 'compute_index_std', 'compute_combinded_std', 'compute_index_fluctuation', 'pl_dvec',
           'compute_index_diff', 'compute_indices', 'filter_indices']

# %% ../../notebooks/detection/01_variance.ipynb 3
import polars as pl
import polars.selectors as cs
from datetime import timedelta

# %% ../../notebooks/detection/01_variance.ipynb 4
from beforerr.polars import pl_norm

def compute_std(
    df: pl.LazyFrame,
    period: timedelta,  # period to group by
    index_column="time",
    cols: list[str] = ["BX", "BY", "BZ"],
    every: timedelta = None,  # every to group by (default: period / 2)
    result_column="std",
):
    if every is None:
        every = period / 2

    std_cols = [col_name + "_std" for col_name in cols]

    std_df = (
        df.group_by_dynamic(index_column, every=every, period=period)
        .agg(
            pl.len(),
            pl.col(cols).std(ddof=0).name.suffix("_std"),
        )
        .with_columns(
            pl_norm(std_cols).alias(result_column),
        )
        .drop(std_cols)
    )
    return std_df

# %% ../../notebooks/detection/01_variance.ipynb 5
def add_neighbor_std(
    df: pl.LazyFrame,
    tau: timedelta,
    join_strategy="inner",
    std_column="std",
    time_column="time",
):
    """
    Get the neighbor standard deviations

    Parameters
    ----------
    - df (pl.LazyFrame): The input DataFrame.
    - tau : The time interval value.

    Notes
    -----
    Simply shift would not work correctly if data is missing, like `std_next = pl.col("B_std").shift(-2)`.

    """

    # Calculate the standard deviation index
    prev_std_df = df.select(
        pl.col(time_column) + tau,
        cs.by_name(std_column, "len").name.suffix("_prev"),
    )

    next_std_df = df.select(
        pl.col(time_column) - tau,
        cs.by_name(std_column, "len").name.suffix("_next"),
    )

    return df.join(prev_std_df, on=time_column, how=join_strategy).join(
        next_std_df, on=time_column, how=join_strategy
    )

# %% ../../notebooks/detection/01_variance.ipynb 6
def compute_index_std(
    df: pl.LazyFrame,
    std_column="std",
):
    """
    Compute the standard deviation index based on the given DataFrame

    Parameters
    ----------
    - df (pl.LazyFrame): The input DataFrame.

    Returns
    -------
    - pl.LazyFrame: DataFrame with calculated 'index_std' column.

    Examples
    --------
    >>> index_std_df = compute_index_std_pl(df)
    >>> index_std_df
    """

    return df.with_columns(
        index_std=pl.col(std_column)
        / pl.max_horizontal(f"{std_column}_prev", f"{std_column}_next"),
    )

# %% ../../notebooks/detection/01_variance.ipynb 8
def compute_combinded_std(
    df: pl.LazyFrame,
    cols: list[str],
    every: timedelta,  # every to group by (default: period / 2)
    period: timedelta = None,  # period to group by
    index_column="time",
    result_column="std_combined",
):
    prev_df = df.with_columns(pl.col(index_column) + period)
    next_df = df.with_columns(pl.col(index_column) - period)
    return (
        pl.concat([prev_df, next_df])
        .sort(index_column)
        .group_by_dynamic(index_column, every=every, period=period)
        .agg(cs.by_name(cols).std(ddof=0).name.suffix("_combined"))
        .select(
            index_column,
            pl_norm([col_name + "_combined" for col_name in cols]).alias(result_column),
        )
    )


# | export
def compute_index_fluctuation(df: pl.LazyFrame, std_column="std", clean=True):
    std_combined = pl.col(f"{std_column}_combined")
    std_added = pl.sum_horizontal(f"{std_column}_prev", f"{std_column}_next")

    index_df = df.with_columns(index_fluctuation=std_combined / std_added)
    if clean:
        return index_df.drop(f"{std_column}_combined")

# %% ../../notebooks/detection/01_variance.ipynb 10
from ..utils.basic import _expand_selectors


def pl_dvec(columns, *more_columns):
    all_columns = _expand_selectors(columns, *more_columns)
    return [
        (pl.col(column).first() - pl.col(column).last()).alias(f"d{column}_vec")
        for column in all_columns
    ]

# %% ../../notebooks/detection/01_variance.ipynb 11
def compute_index_diff(
    df: pl.LazyFrame,
    every: timedelta,
    cols: list[str],
    period: timedelta = None,
    clean=True,
):
    db_cols = ["d" + col + "_vec" for col in cols]

    index_diff = (
        df.with_columns(pl_norm(cols).alias("_vec_mag"))
        .group_by_dynamic("time", every=every, period=period)
        .agg(
            pl.col("_vec_mag").mean().name.suffix("_mean"),
            *pl_dvec(cols),
        )
        .with_columns(pl_norm(db_cols).alias("_dvec_mag"))
        .with_columns(index_diff=pl.col("_dvec_mag") / pl.col("_vec_mag_mean"))
    )

    if clean:
        return index_diff.drop("_vec_mag", "_vec_mag_mean", "_dvec_mag", *db_cols)
    else:
        return index_diff

# %% ../../notebooks/detection/01_variance.ipynb 12
def compute_indices(
    df: pl.LazyFrame,
    tau: timedelta,
    cols: list[str] = ["BX", "BY", "BZ"],
    clean=True,
    join_strategy="inner",
) -> pl.LazyFrame:
    """
    Compute all index based on the given DataFrame and tau value.

    Parameters
    ----------
    df : pl.DataFrame
        Input DataFrame.
    tau : datetime.timedelta
        Time interval value.

    Returns
    -------
    tuple :
        Tuple containing DataFrame results for fluctuation index,
        standard deviation index, and 'index_num'.

    Examples
    --------
    >>> indices = compute_indices(df, tau)

    Notes
    -----
    - This is a wrapper for `_compute_indices` with `pl.LazyFrame` input.
    - Simply shift to calculate index_std would not work correctly if data is missing,
        like `std_next = pl.col("B_std").shift(-2)`.
    - Drop null though may lose some IDs (using the default `join_strategy`).
        Because we could not tell if it is a real ID or just a partial wave
        from incomplete data without previous or/and next std.
        Hopefully we can pick up the lost ones with smaller tau.
    - TODO: Can be optimized further, but this is already fast enough.
        - TEST: if `join` can be improved by shift after filling the missing values.
        - TEST: if `list` in `polars` really fast?
    """

    every = tau / 2
    period = tau

    stds_df = df.pipe(compute_std, period=period, cols=cols).pipe(
        add_neighbor_std, tau=tau
    )

    combined_std_df = compute_combinded_std(df, cols, every=every, period=period)

    indices = (
        df.pipe(compute_index_diff, every=every, period=period, cols=cols)
        .join(stds_df, on="time")
        .join(combined_std_df, on="time", how=join_strategy)
        .pipe(compute_index_std)
        .pipe(compute_index_fluctuation)
    )

    if clean:
        return indices.drop(
            ["B_std_prev", "B_std_next", "B_added_std", "B_std_combined"]
        )
    else:
        return indices

# %% ../../notebooks/detection/01_variance.ipynb 14
INDEX_STD_THRESHOLD = 2
INDEX_FLUC_THRESHOLD = 1
INDEX_DIFF_THRESHOLD = 0.1
SPARSE_THRESHOLD = 15

# %% ../../notebooks/detection/01_variance.ipynb 15
def filter_indices(
    df: pl.LazyFrame,
    index_std_threshold: float = INDEX_STD_THRESHOLD,
    index_fluc_threshold: float = INDEX_FLUC_THRESHOLD,
    index_diff_threshold: float = INDEX_DIFF_THRESHOLD,
    sparse_num: int = SPARSE_THRESHOLD,
) -> pl.LazyFrame:
    # filter indices to get possible IDs

    return df.filter(
        pl.col("index_std") > index_std_threshold,
        pl.col("index_fluctuation") > index_fluc_threshold,
        pl.col("index_diff") > index_diff_threshold,
        pl.col(
            "index_std"
        ).is_finite(),  # for cases where neighboring groups have std=0
        pl.col("len") > sparse_num,
        pl.col("len_prev")
        > sparse_num,  # filter out sparse intervals, which may give unreasonable results.
        pl.col("len_next")
        > sparse_num,  # filter out sparse intervals, which may give unreasonable results.
    ).drop(["len_prev", "len_next"])



__version__ = "0.1.0"
# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/__init__.ipynb.

# %% auto 0
__all__ = []

# %% ../notebooks/__init__.ipynb 1
from .datasets import IdsEvents, IDsDataset



# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/10_datasets.ipynb.

# %% auto 0
__all__ = ['write', 'IdsEvents', 'log_event_change', 'IDsDataset']

# %% ../notebooks/10_datasets.ipynb 1
import polars as pl
# import holoviews as hv
import polars.selectors as cs
from loguru import logger
from random import sample
from datetime import timedelta

from pydantic import BaseModel, Field, validate_call
from space_analysis.ds.meta import Meta, PlasmaMeta, TempMeta
from typing import Literal

# %% ../notebooks/10_datasets.ipynb 3
from .utils.basic import df2ts
from .utils.plot import plot_candidate as _plot_candidate
from .integration import combine_features, calc_combined_features
from .core.pipeline import ids_finder

# %% ../notebooks/10_datasets.ipynb 4
from pathlib import Path


def write(df: pl.DataFrame, fname: Path, format=None, **kwargs):
    if format is None:
        format = fname.suffix
        format = format[1:] if format.startswith(".") else format
    match format:
        case "arrow":
            df.write_ipc(fname, **kwargs)
        case "csv":
            df.write_csv(fname, **kwargs)
        case "parquet":
            df.write_parquet(fname, **kwargs)

    logger.info(f"Dataframe written to {fname}")

    return fname


class IdsEvents(BaseModel):
    """Core class to handle discontinuity events in a dataset."""

    class Config:
        extra = "allow"
        arbitrary_types_allowed = True

    name: str = None
    data: pl.LazyFrame = None
    ts: timedelta = None
    """time resolution of the dataset"""
    tau: timedelta = None
    """time interval used to find events"""
    events: pl.DataFrame = None
    method: Literal["fit", "derivative"] = "fit"

    @validate_call
    def export(self, path: Path, format="arrow", clean=True, **kwargs):
        if self.events is None:
            self.find_events()
        _df = self.events
        if clean:
            _df = _df.select(cs.datetime(), cs.duration(), cs.numeric())

        # check the parent directory exists
        path.parent.mkdir(parents=True, exist_ok=True)
        write(_df, path, format, **kwargs)
        return self

    def find_events(self, **kwargs):
        self.events = ids_finder(
            self.data, ts=self.ts, tau=self.tau, method=self.method, **kwargs
        )
        return self

    def get_event(self, index: int):
        events = self.events
        if "index" not in events.columns:
            events = events.with_row_index()
        predicate = pl.col("index") == index
        return events.row(by_predicate=predicate, named=True)

    def get_events(self, indices=None, predicate=None):
        pass

    def get_event_data(
        self,
        event,
        start_col="t.d_start",
        end_col="t.d_end",
        offset=timedelta(seconds=1),
        **kwargs,
    ):
        start = event[start_col] - offset
        end = event[end_col] + offset

        _data = self.data.filter(pl.col("time").is_between(start, end))
        return df2ts(_data, self.bcols)

# %% ../notebooks/10_datasets.ipynb 5
def log_event_change(event, logger=logger):
    logger.debug(
        f"""CHANGE INFO
        n.change: {event.get('n.change')}
        v.ion.change: {event.get('v.ion.change')}
        T.change: {event.get('T.change')}
        v.Alfven.change: {event.get('v.Alfven.change')}
        v.ion.change.l: {event.get('v.ion.change.l')}
        v.Alfven.change.l: {event.get('v.Alfven.change.l')}
        """
    )

# %% ../notebooks/10_datasets.ipynb 8
class IDsDataset(IdsEvents):
    """Extend the IdsEvents class to handle plasma and temperature data."""

    data: pl.LazyFrame = Field(default=None, alias="mag_data")
    mag_meta: Meta = Meta()
    bcols: list[str] = None

    plasma_data: pl.LazyFrame = None
    plasma_meta: PlasmaMeta = PlasmaMeta()

    ion_temp_data: pl.LazyFrame = None
    ion_temp_meta: TempMeta = TempMeta()
    e_temp_data: pl.LazyFrame = None
    e_temp_meta: TempMeta = TempMeta()

    def update_events(self, **kwargs):
        return self.update_events_with_plasma_data(
            **kwargs
        ).update_events_with_temp_data(**kwargs)

    def update_events_with_plasma_data(self, **kwargs):
        # TypeError: the truth value of a LazyFrame is ambiguous
        if self.plasma_data is not None:
            df_combined = combine_features(
                self.events,
                self.plasma_data.collect(),
                plasma_meta=self.plasma_meta,
                **kwargs,
            )

            self.events = calc_combined_features(
                df_combined,
                plasma_meta=self.plasma_meta,
                **kwargs,
            )
        else:
            logger.info("Plasma data is not available.")

        return self

    def update_events_with_temp_data(self, **kwargs):
        on = "time"

        if self.ion_temp_data is not None:
            self.events = self.events.sort(on).join_asof(
                self.ion_temp_data.sort(on).collect(), on=on
            )
        else:
            logger.info("Ion temperature data is not available.")

        if self.e_temp_data is not None:
            self.events = self.events.sort(on).join_asof(
                self.e_temp_data.sort(on).collect(), on=on
            )
        else:
            logger.info("Electron temperature data is not available.")
        return self

    def plot(self, type="overview", event=None, index=None, **kwargs):

        event = event or self.get_event(index)
        if type == "overview":
            return self.overview_plot(event, **kwargs)

    def plot_candidate(self, event=None, index=None, **kwargs):
        if event is None:
            event = self.get_event(index)
        data = self.get_event_data(event, **kwargs)

        return _plot_candidate(event, data, **kwargs)

    def plot_candidates(
        self, indices=None, num=4, random=True, predicate=None, **kwargs
    ):
        events = self.events
        if "index" not in events.columns:
            events = events.with_row_index()

        if indices is None:  # the truth value of an Expr is ambiguous
            if predicate is not None:
                events = events.filter(predicate)
            indices = events.get_column("index")
            if random:
                indices = indices.sample(num).to_numpy()
            else:
                indices = indices.head(num).to_numpy()
            logger.info(f"Candidates indices: {indices}")

        return [self.plot_candidate(index=i, **kwargs) for i in indices]



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/01_ids_detection.ipynb.

# %% auto 0
__all__ = ['pl_format_time', 'detect_events']

# %% ../../notebooks/01_ids_detection.ipynb 1
from datetime import timedelta
import polars as pl

from typing import overload

# %% ../../notebooks/01_ids_detection.ipynb 2
# some helper functions
@overload
def pl_format_time(df: pl.DataFrame, tau: timedelta) -> pl.DataFrame: pass

@overload
def pl_format_time(df: pl.LazyFrame, tau: timedelta) -> pl.LazyFrame: pass

def pl_format_time(df: pl.LazyFrame, tau: timedelta):
    return df.with_columns(
        tstart=pl.col("time"),
        tstop=(pl.col("time") + tau),
        time=(pl.col("time") + tau / 2),
    )

# %% ../../notebooks/01_ids_detection.ipynb 4
from ..detection.variance import compute_indices, filter_indices

def detect_events(
    data: pl.LazyFrame,
    tau: timedelta,
    ts: timedelta,
    bcols,
    sparse_num=None,
    method="liu",
    **kwargs,
):
    indices = compute_indices(data, tau, bcols)
    if sparse_num is None:
        sparse_num = tau / ts // 3

    events = indices.pipe(filter_indices, sparse_num=sparse_num).pipe(
        pl_format_time, tau
    ).collect()

    return events



# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/12_mission_config.ipynb.

# %% auto 0
__all__ = ['wind_plasma_3dp_meta', 'wind_plasma_k0_swe_meta', 'wind_mag_h2_meta', 'wind_mag_h4_rtn_meta', 'WindMeta',
           'WindConfig']

# %% ../notebooks/12_mission_config.ipynb 0
from space_analysis.ds.meta import Meta, PlasmaMeta, TempMeta
from .config import SpeasyIDsConfig
from datetime import timedelta
from pydantic import BaseModel

# %% ../notebooks/12_mission_config.ipynb 1
wind_plasma_3dp_meta = PlasmaMeta(
    dataset="WI_PM_3DP",
    parameters=["P_DENS", "P_VELS", "P_TEMP"],  # Proton velocity vector (GSE)
    description = "Wind 3dp, PESA LOW 1 spin resolution ion (proton and alpha) moments (computed on spacecraft)"
)

wind_plasma_k0_swe_meta = PlasmaMeta(
    dataset="WI_K0_SWE", parameters=["Np", "V_GSE", "THERMAL_SPD"]
)

wind_mag_h2_meta = Meta(dataset="WI_H2_MFI", parameters=["BGSE"])
wind_mag_h4_rtn_meta: Meta = Meta(dataset="WI_H4-RTN_MFI", parameters=["BRTN"])

# %% ../notebooks/12_mission_config.ipynb 2
class WindMeta(BaseModel):
    name: str = "Wind"
    ts: timedelta = timedelta(seconds=1 / 11)

    mag_meta: Meta = wind_mag_h2_meta
    plasma_meta: PlasmaMeta = wind_plasma_3dp_meta

    ion_temp_meta: TempMeta = TempMeta(
        dataset="WI_PLSP_3DP",
        parameters=["MOM.P.MAGT3"],  # T_perp = (MAGT3[0] + MAGT3[1])/2
        para_col="proton_MagT3_Z",
        perp_cols=["proton_MagT3_X", "proton_MagT3_Y"],
    )

    e_temp_meta: TempMeta = TempMeta(
        dataset="WI_ELM2_3DP",
        parameters=["MAGT3"],
        para_col="electron_MagT3_Para",
        perp_cols=["electron_MagT3_Perp1", "electron_MagT3_Perp2"],
    )

# %% ../notebooks/12_mission_config.ipynb 3
class WindConfig(WindMeta, SpeasyIDsConfig):
    pass



# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/21_datasets_kedro.ipynb.

# %% auto 0
__all__ = ['E1IDsDataset', 'cIDsDataset']

# %% ../notebooks/21_datasets_kedro.ipynb 1
from .datasets import IDsDataset

from datetime import timedelta

import polars as pl
import polars.selectors as cs

from kedro.io import DataCatalog
from .utils.basic import concat_partitions

# %% ../notebooks/21_datasets_kedro.ipynb 2
class E1IDsDataset(IDsDataset):
    sat_id: str
    tau: timedelta
    ts: timedelta = timedelta(seconds=1)

# %% ../notebooks/21_datasets_kedro.ipynb 3
class cIDsDataset(E1IDsDataset):
    catalog: DataCatalog = None

    _load_data_format = "{sat}.MAG.primary_data_{ts}"
    _load_events_format = "events.{sat}_{ts}_{tau}"
    or_df: pl.DataFrame | None = None  # occurence rate
    or_df_normalized: pl.DataFrame | None = None  # normalized occurence rate

    def __init__(self, **data):
        super().__init__(**data)

        tau_str = f"tau_{self.tau.seconds}s"
        ts_mag_str = f"ts_{self.ts.seconds}s"

        self._tau_str = tau_str
        self._ts_mag_str = ts_mag_str

        self.events_format = self._load_events_format.format(
            sat=self.sat_id, ts=ts_mag_str, tau=tau_str
        )

        if data.get("data_format") is None:
            self.data_format = self._load_data_format.format(
                sat=self.sat_id, ts=ts_mag_str
            )

        if self.candidates is None:
            self.load_events()
        if self.mag_data is None:
            self.load_data()

    def load_events(self):
        data_format = self.events_format
        self.candidates = (
            self.catalog.load(data_format)
            .fill_nan(None)
            .with_columns(
                cs.float().cast(pl.Float64),
                sat=pl.lit(self.sat_id),
            )
            .collect()
        )

    def load_data(self):
        data_format = self.data_format
        self.mag_data = concat_partitions(self.catalog.load(data_format))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/properties/00_duration.ipynb.

# %% auto 0
__all__ = ['THRESHOLD_RATIO', 'AvailableMethod', 'ts_max_distance', 'ts_max_derivative', 'find_start_end_times',
           'get_time_from_condition', 'calc_duration']

# %% ../../notebooks/properties/00_duration.ipynb 1
from scipy.spatial import distance_matrix
import xarray as xr
import numpy as np
import pandas as pd
from xarray_einstats import linalg
from typing import Literal

# %% ../../notebooks/properties/00_duration.ipynb 3
def ts_max_distance(
    ts: xr.DataArray, coord: str = "time"
):
    "Compute the time interval when the timeseries has maxium cumulative variation"
    distance = distance_matrix(ts.data, ts.data)
    max_distance_index = np.unravel_index(np.argmax(distance), distance.shape)
    return ts[coord].values[list(max_distance_index)]

# %% ../../notebooks/properties/00_duration.ipynb 6
THRESHOLD_RATIO = 1 / 4


def ts_max_derivative(vec: xr.DataArray, threshold_ratio=THRESHOLD_RATIO):
    # NOTE: gradient calculated at the edge is not reliable.
    vec_diff = vec.differentiate("time", datetime_unit="s").isel(time=slice(1, -1))
    vec_diff_mag = linalg.norm(vec_diff, dims="v_dim")

    # Determine d_star based on trend
    if vec_diff_mag.isnull().all():
        raise ValueError(
            "The differentiated vector magnitude contains only NaN values. Cannot compute duration."
        )

    d_star_index = vec_diff_mag.argmax(dim="time")
    d_star = vec_diff_mag[d_star_index].item()
    d_time = vec_diff_mag.time[d_star_index].values

    threshold = d_star * threshold_ratio

    start_time, end_time = find_start_end_times(vec_diff_mag, d_time, threshold)

    return start_time, end_time, d_time, d_star


def find_start_end_times(
    vec_diff_mag: xr.DataArray, d_time, threshold
) -> tuple[pd.Timestamp, pd.Timestamp]:
    # Determine start time
    pre_vec_mag = vec_diff_mag.sel(time=slice(None, d_time))
    start_time = get_time_from_condition(pre_vec_mag, threshold, "last_below")

    # Determine stop time
    post_vec_mag = vec_diff_mag.sel(time=slice(d_time, None))
    end_time = get_time_from_condition(post_vec_mag, threshold, "first_below")

    return start_time, end_time


def get_time_from_condition(
    vec: xr.DataArray, threshold, condition_type
) -> pd.Timestamp:
    if condition_type == "first_below":
        condition = vec < threshold
        index_choice = 0
    elif condition_type == "last_below":
        condition = vec < threshold
        index_choice = -1
    else:
        raise ValueError(f"Unknown condition_type: {condition_type}")

    where_result = np.where(condition)[0]

    if len(where_result) > 0:
        return vec.time[where_result[index_choice]].values
    return None

# %% ../../notebooks/properties/00_duration.ipynb 7
AvailableMethod = Literal["distance", "derivative"]

def calc_duration(ts: xr.DataArray, method: AvailableMethod = "distance", **kwargs):
    if method == "distance":
        result = np.sort(ts_max_distance(ts))
        keys = ["t.d_start", "t.d_end"]

    elif method == "derivative":
        result = ts_max_derivative(ts)
        keys = ["t.d_start", "t.d_end", "t.d_time", "d_star_max"]

    return dict(zip(keys, result))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/11_ids_config.ipynb.

# %% auto 0
__all__ = ['standardize_plasma_data', 'IDsConfig', 'SpeasyIDsConfig']

# %% ../notebooks/11_ids_config.ipynb 0
from datetime import datetime
from .datasets import IDsDataset
from space_analysis.ds.meta import Meta, PlasmaMeta
from space_analysis.utils.speasy import Variables
import polars as pl
from loguru import logger
from pathlib import Path
from functools import cached_property

from tqdm.auto import tqdm

# %% ../notebooks/11_ids_config.ipynb 1
def standardize_plasma_data(data: pl.LazyFrame, meta: PlasmaMeta):
    """
    Standardize plasma data columns across different datasets.
    
    Notes: meta will be updated with the new column names
    """

    if meta.density_col:
        data = data.rename({meta.density_col: "plasma_density"})
        meta.density_col = "plasma_density"
    return data

# %% ../notebooks/11_ids_config.ipynb 2
class IDsConfig(IDsDataset):
    """
    Extend the IDsDataset class to provide additional functionalities:
    
    - Export and load data
    - Standardize data
    - Split data to handle large datasets
    """
    timerange: list[datetime] = None

    split: int = 1
    fmt: str = "arrow"
    
    _data_dir = Path("data")
    
    @property
    def fname(self):
        ts_str = f"ts_{self.ts.total_seconds():.2f}s"
        tau_str = f"tau_{self.tau.total_seconds():.0f}s"
        return f"events.{self.name}.{self.method}.{ts_str}_{tau_str}.{self.fmt}"

    @property
    def path(self):
        return self._data_dir / self.fname
    
    @property
    def timeranges(self):
        from sunpy.time import TimeRange

        trs: list[TimeRange] = TimeRange(self.timerange).split(self.split)
        return [[tr.start.value, tr.end.value] for tr in trs]
    
    def export(self, **kwargs):
        return super().export(self.path, format=self.fmt, **kwargs)

    def load(self):
        if self.path.exists():
            logger.info(f"Loading data from {self.path}")
            self.events = pl.read_ipc(self.path)
        else:
            logger.warning(f"Data not found at {self.path}")
        return self

# %% ../notebooks/11_ids_config.ipynb 3
class SpeasyIDsConfig(IDsConfig):
    """Based on `speasy` Variables to get the data"""
    provider: str = "cda"

    def model_post_init(self, __context):
        # TODO: directly get columns from the data without loading them
        # self.plasma_meta.density_col = self.plasma_vars.data[0].columns[0]
        # self.plasma_meta.velocity_cols = self.plasma_vars.data[1].columns
        pass

    def get_vars(self, vars: str):
        meta: Meta = getattr(self, f"{vars}_meta")
        return Variables(
            timerange=self.timerange,
            provider=self.provider,
            **meta.model_dump(exclude_unset=True),
        )

    def get_vars_df(self, vars: str, cached: bool = False):
        if not cached:
            return self.get_vars(vars).to_polars()
        else:
            return NotImplementedError

    # Variables
    @cached_property
    def mag_vars(self):
        return self.get_vars("mag")

    @cached_property
    def plasma_vars(self):
        return self.get_vars("plasma")

    @cached_property
    def ion_temp_var(self):
        return self.get_vars("ion_temp")

    @cached_property
    def e_temp_var(self):
        return self.get_vars("e_temp")

    # DataFrames
    def set_data_from_vars(self, update: False):
        pass

    def _get_and_process_data(self, **kwargs):
        self.plasma_meta.density_col = self.plasma_vars.data[0].columns[0]
        self.plasma_meta.velocity_cols = self.plasma_vars.data[1].columns

        # TODO: optimize for no-split timeranges

        for tr in tqdm(self.timeranges):
            ids_ds = self.model_copy(update={"timerange": tr, "split": 1}, deep=True)

            ids_ds.data = ids_ds.get_vars_df("mag", cached=False)
            ids_ds.plasma_data = ids_ds.get_vars_df("plasma", cached=False).pipe(
                standardize_plasma_data, ids_ds.plasma_meta
            )

            yield ids_ds.find_events(
                return_best_fit=False
            ).update_events_with_plasma_data().events
            
    def get_and_process_data(self, **kwargs):
        self.events = pl.concat(self._get_and_process_data(**kwargs))
        return self



# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/03_mag_plasma.ipynb.

# %% auto 0
__all__ = ['J_FACTOR', 'interpolate', 'interpolate2', 'format_time', 'combine_features', 'vector_project', 'vector_project_pl',
           'calc_rotation_angle_pl', 'compute_inertial_length', 'compute_Alfven_current',
           'calc_plasma_parameter_change', 'calc_combined_features']

# %% ../notebooks/03_mag_plasma.ipynb 1
import polars as pl
import polars.selectors as cs
from space_analysis.plasma.formulary import df_Alfven_speed as compute_Alfven_speed

# %% ../notebooks/03_mag_plasma.ipynb 2
def interpolate(
    df: pl.DataFrame, on="time", method="index", limit=1, limit_direction="both"
):
    # Note: limit is set to 1 to improve the confidence of the interpolation
    return pl.from_pandas(
        df.to_pandas()
        .set_index(on)
        .sort_index()
        .interpolate(
            method=method,
            limit=limit,
            limit_direction=limit_direction,
        )
        .reset_index()
    )


def interpolate2(df1: pl.DataFrame, df2, **kwargs):
    return pl.concat([df1, df2], how="diagonal_relaxed").pipe(interpolate, **kwargs)


def format_time(df: pl.DataFrame, time_unit="ns"):
    return df.with_columns(
        cs.datetime().dt.cast_time_unit(time_unit),
    )

# %% ../notebooks/03_mag_plasma.ipynb 3
from space_analysis.ds.meta import PlasmaMeta
from fastcore.all import concat


def combine_features(
    events: pl.DataFrame,
    states_data: pl.DataFrame,
    plasma_meta: PlasmaMeta = PlasmaMeta(),
    method: str = "interpolate",
):

    m = plasma_meta
    subset_cols = concat([m.density_col, m.velocity_cols, m.temperature_col])
    subset_cols = [item for item in subset_cols if item is not None]

    # change time format: see issue: https://github.com/pola-rs/polars/issues/12023
    time_unit = events["time"].dtype.time_unit
    states_data = states_data.sort("time").with_columns(
        pl.col("time").dt.cast_time_unit(time_unit)
    )
    events = events.sort("time")

    subset_cols = subset_cols + [
        "time"
    ]  # https://stackoverflow.com/questions/2347265/why-does-behave-unexpectedly-on-lists
    states_data_subset = states_data.select(subset_cols)
    df = events.join_asof(states_data, on="time", strategy="nearest")

    if method == "interpolate":
        before_df = interpolate2(
            df.select(time=pl.col("t.d_start")), states_data_subset
        )
        after_df = interpolate2(df.select(time=pl.col("t.d_end")), states_data_subset)
        return (
            df.sort("t.d_start")
            .join(
                before_df,
                left_on="t.d_start",
                right_on="time",
                suffix="_before",
            )
            .sort("t.d_end")
            .join(
                after_df,
                left_on="t.d_end",
                right_on="time",
                suffix="_after",
            )
        )

    elif method == "nearest":
        return (
            df.sort("t.d_start")
            .join_asof(
                states_data_subset,
                left_on="t.d_start",
                right_on="time",
                strategy="backward",
                suffix="_before",
            )
            .sort("t.d_end")
            .join_asof(
                states_data_subset,
                left_on="t.d_end",
                right_on="time",
                strategy="forward",
                suffix="_after",
            )
        )
    else:
        return df

# %% ../notebooks/03_mag_plasma.ipynb 5
import astropy.units as u
from astropy.constants import mu0, e
from plasmapy.formulary.lengths import inertial_length
from xarray_einstats import linalg
from .utils.basic import df2ts
from .core.propeties import calc_rotation_angle
import xarray as xr

# %% ../notebooks/03_mag_plasma.ipynb 6
def vector_project(v1: xr.DataArray, v2: xr.DataArray, dim="v_dim"):
    return xr.dot(v1, v2, dims=dim) / linalg.norm(v2, dims=dim)


def vector_project_pl(df: pl.DataFrame, v1_cols, v2_cols, name=None):
    v1 = df2ts(df, v1_cols).assign_coords(v_dim=["x", "y", "z"])
    v2 = df2ts(df, v2_cols).assign_coords(v_dim=["x", "y", "z"])
    result = vector_project(v1, v2, dim="v_dim")

    return df.with_columns(pl.Series(result.data).alias(name or "v_proj"))

# %% ../notebooks/03_mag_plasma.ipynb 8
def calc_rotation_angle_pl(df: pl.DataFrame, v1_cols, v2_cols, name):
    v1 = df.select(v1_cols).to_numpy()
    v2 = df.select(v2_cols).to_numpy()

    result = calc_rotation_angle(v1, v2)

    return df.with_columns(pl.Series(result).alias(name))

# %% ../notebooks/03_mag_plasma.ipynb 10
def compute_inertial_length(df: pl.DataFrame, density_col="plasma_density"):
    density = df[density_col].to_numpy() * u.cm ** (-3)
    result = inertial_length(density, "H+").to(u.km)

    return df.with_columns(ion_inertial_length=pl.Series(result.value))

# %% ../notebooks/03_mag_plasma.ipynb 12
def compute_Alfven_current(
    df: pl.DataFrame,
    density_col="plasma_density",
):
    Alfven_speed = df["Alfven_speed"].to_numpy() * u.km / u.s
    density = df[density_col].to_numpy() * u.cm ** (-3)

    result = (e.si * Alfven_speed * density).to(u.nA / u.m**2)

    return df.with_columns(j_Alfven=pl.Series(result.value))

# %% ../notebooks/03_mag_plasma.ipynb 14
from space_analysis.ds.meta import PlasmaMeta


def calc_plasma_parameter_change(
    df: pl.DataFrame,
    plasma_meta: PlasmaMeta = PlasmaMeta(),
):
    if plasma_meta.temperature_col:
        col = plasma_meta.temperature_col
        df = df.rename(
            {
                f"{col}_before": "T.before",
                f"{col}_after": "T.after",
            }
        ).with_columns((pl.col("T.after") - pl.col("T.before")).alias("T.change"))

    if plasma_meta.speed_col:
        col = plasma_meta.speed_col
        df = df.rename(
            {
                f"{col}_before": "v.ion.before",
                f"{col}_after": "v.ion.after",
            }
        ).with_columns(
            (pl.col("v.ion.after") - pl.col("v.ion.before")).alias("v.ion.change")
        )

    return (
        df.rename(
            {
                "plasma_density_before": "n.before",
                "plasma_density_after": "n.after",
            }
        )
        .pipe(
            compute_Alfven_speed,
            B="B.before",
            n="n.before",
            col_name="v.Alfven.before",
            sign=True,
        )
        .pipe(
            compute_Alfven_speed,
            B="B.after",
            n="n.after",
            col_name="v.Alfven.after",
            sign=True,
        )
        .pipe(
            compute_Alfven_speed,
            n="n.before",
            B="B.vec.before.l",
            col_name="v.Alfven.before.l",
            sign=True,
        )
        .pipe(
            compute_Alfven_speed,
            n="n.after",
            B="B.vec.after.l",
            col_name="v.Alfven.after.l",
            sign=True,
        )
        .with_columns(
            (pl.col("n.after") - pl.col("n.before")).alias("n.change"),
            (pl.col("v.ion.after.l") - pl.col("v.ion.before.l")).alias(
                "v.ion.change.l"
            ),
            (pl.col("B.after") - pl.col("B.before")).alias("B.change"),
            (pl.col("v.Alfven.after") - pl.col("v.Alfven.before")).alias(
                "v.Alfven.change"
            ),
            (pl.col("v.Alfven.after.l") - pl.col("v.Alfven.before.l")).alias(
                "v.Alfven.change.l"
            ),
        )
    )

# %% ../notebooks/03_mag_plasma.ipynb 15
from beforerr.polars import decompose_vector
from space_analysis.ds.meta import PlasmaMeta

J_FACTOR = ((u.nT / u.s) * (1 / mu0 / (u.km / u.s))).to(u.nA / u.m**2)


def calc_combined_features(
    df: pl.DataFrame,
    b_cols: list[str] = None,  # ["B_background_x", "B_background_y", "B_background_z"]
    detail: bool = True,
    normal_cols: list[str] = ["k_x", "k_y", "k_z"],
    Vl_cols=["Vl_x", "Vl_y", "Vl_z"],
    Vn_cols=["Vn_x", "Vn_y", "Vn_z"],
    thickness_cols=["L_k"],
    current_cols=["j0_k"],
    plasma_meta: PlasmaMeta = None,
    **kwargs,
):
    """Calculate the combined features of the discontinuity

    Args:
        df (pl.DataFrame): _description_
        normal_cols (list[str], optional): normal vector of the discontinuity plane. Defaults to [ "k_x", "k_y", "k_z", ].
        detail (bool, optional): _description_. Defaults to True.
        Vl_cols (list, optional): maxium variance direction vector of the magnetic field. Defaults to [ "Vl_x", "Vl_y", "Vl_z", ].
        Vn_cols (list, optional): minimum variance direction vector of the magnetic field. Defaults to [ "Vn_x", "Vn_y", "Vn_z", ].
        current_cols (list, optional): _description_. Defaults to ["j0_mn", "j0_k"].
    """

    length_norm = pl.col("ion_inertial_length")
    current_norm = pl.col("j_Alfven")
    b_norm = pl.col("b_mag")

    result = df

    if b_cols is not None:
        result = result.pipe(
            calc_rotation_angle_pl, b_cols, normal_cols, name="theta_n_b"
        )

    vec_cols = plasma_meta.velocity_cols
    density_col = plasma_meta.density_col

    result = (
        result.pipe(vector_project_pl, vec_cols, Vl_cols, name="v_l")
        .pipe(vector_project_pl, vec_cols, Vn_cols, name="v_n")
        .pipe(vector_project_pl, vec_cols, normal_cols, name="v_k")
        .with_columns(
            pl.col("v_n").abs(),
            pl.col("v_k").abs(),
            # v_mn=(pl.col("plasma_speed") ** 2 - pl.col("v_l") ** 2).sqrt(),
        )
        .with_columns(
            L_k=pl.col("v_k") * pl.col("duration"),
            j0_k=pl.col("d_star")
            / pl.col(
                "v_k"
            ),  # TODO: d_star corresponding to dB/dt, which direction is not exactly perpendicular to the k direction
            # NOTE: n direction is not properly determined for MVA analysis
            # j0_mn=pl.col("d_star") / pl.col("v_mn"),
            # L_n=pl.col("v_n") * pl.col("duration"),
            # L_mn=pl.col("v_mn") * pl.col("duration"),
            # NOTE: the duration is not properly determined for `max distance` method
            # L_k=pl.col("v_k") * pl.col("duration"),
        )
        .pipe(compute_inertial_length)
        .pipe(compute_Alfven_speed, n=density_col, B="b_mag")
        .pipe(compute_Alfven_current)
        .with_columns(
            cs.by_name(current_cols) * J_FACTOR.value,
        )
        .with_columns(
            (cs.by_name(thickness_cols) / length_norm).name.suffix("_norm"),
            (cs.by_name(current_cols) / current_norm).name.suffix("_norm"),
            (cs.by_name(b_cols) / b_norm).name.suffix("_norm"),
        )
    )

    if detail:
        result = (
            result.pipe(
                vector_project_pl,
                [_ + "_before" for _ in vec_cols],
                Vl_cols,
                name="v.ion.before.l",
            )
            .pipe(
                vector_project_pl,
                [_ + "_after" for _ in vec_cols],
                Vl_cols,
                name="v.ion.after.l",
            )
            .pipe(decompose_vector, "B.vec.before", suffixes=[".l", ".m", ".n"])
            .pipe(decompose_vector, "B.vec.after", suffixes=[".l", ".m", ".n"])
            .pipe(calc_plasma_parameter_change, plasma_meta=plasma_meta)
        )

    return result
